--------------------------------------------------------------------------------
-- Part 1: Two Languages with Natural Numbers
--------------------------------------------------------------------------------

module Exam.Part1.Properties where

open import Data.Nat using (ℕ; suc; zero; _+_)
open import Relation.Binary.PropositionalEquality using (_≡_; refl; sym; trans; cong; cong₂; module ≡-Reasoning)

open import Exam.Part1.Specification

--------------------------------------------------------------------------------

-- Evaluating an expression in one language is the same as translating
-- it to the other language and evaluating it there.

sound : ∀ (e : Expr) → ⟦ e ⟧ ≡ ⟦ tr e ⟧'
sound = {!!}

complete : ∀ (e : Expr') → ⟦ e ⟧' ≡ ⟦ tr' e ⟧
complete = {!!}

-- Translating an expression back and forth between the languages
-- doesn't change its meaning.

tr'-tr : ∀ (e : Expr) → ⟦ tr' (tr e) ⟧ ≡ ⟦ e ⟧
tr'-tr e = {!!}

tr-tr' : ∀ (e : Expr') → ⟦ tr (tr' e) ⟧' ≡ ⟦ e ⟧'
tr-tr' e = {!!}

--------------------------------------------------------------------------------

-- If one expression reduces to another, then they have the same denotation.

completeness-↪ : ∀ {e₁ e₂ : Expr} → e₁ ↪ e₂ → ⟦ e₁ ⟧ ≡ ⟦ e₂ ⟧
completeness-↪ = {!!}

completeness-↪* : ∀ {e₁ e₂ : Expr} → e₁ ↪* e₂ → ⟦ e₁ ⟧ ≡ ⟦ e₂ ⟧
completeness-↪* = {!!}

completeness-↪' : ∀ {e₁ e₂ : Expr'} → e₁ ↪' e₂ → ⟦ e₁ ⟧' ≡ ⟦ e₂ ⟧'
completeness-↪' = {!!}

completeness-↪'* : ∀ {e₁ e₂ : Expr'} → e₁ ↪'* e₂ → ⟦ e₁ ⟧' ≡ ⟦ e₂ ⟧'
completeness-↪'* = {!!}

--------------------------------------------------------------------------------

-- All expressions reduce in zero or more steps to their denotation.
--
-- Hint: use the ↪*-Reasoning and ↪*'-Reasoning modules here to have
-- more readable proofs. See the recording of the last tutorial session
-- if you're not sure how to use multiple reasoning chains in one file.
--
-- Hint: You need to prove a few simple lemmas, which look like
-- the ξ-rules but do zero or more steps instead of exactly one.

-- Lemmas

ξ*-+₁ : ∀ {e₁ e₂ e₁' : Expr} → e₁ ↪* e₁' → (e₁ `+ e₂) ↪* (e₁' `+ e₂)
ξ*-+₁ = {!!}

ξ*-+₂ : ∀ {e₁ e₂ e₂' : Expr} → e₂ ↪* e₂' → (e₁ `+ e₂) ↪* (e₁ `+ e₂')
ξ*-+₂ = {!!}

-- What we want to prove

↪-⟦⟧ : ∀ (e : Expr) → e ↪* (# ⟦ e ⟧)
↪-⟦⟧ = {!!}

-- Lemmas

ξ*-+₁' : ∀ {e₁ e₂ e₁' : Expr'} → e₁ ↪'* e₁' → (e₁ `+ e₂) ↪'* (e₁' `+ e₂)
ξ*-+₁' = {!!}

ξ*-+₂' : ∀ {e₁ e₂ e₂' : Expr'} → e₂ ↪'* e₂' → (e₁ `+ e₂) ↪'* (e₁ `+ e₂')
ξ*-+₂' = {!!}

ξ*-suc' : ∀ {e e' : Expr'} → e ↪'* e' → (`suc e) ↪'* (`suc e')
ξ*-suc' = {!!}

β-+-lem : ∀ (n₁ n₂ : ℕ) → (ℕ→Expr' n₁ `+ ℕ→Expr' n₂) ↪'* (ℕ→Expr' (n₁ + n₂))
β-+-lem = {!!}

-- What we want to prove

↪'-⟦⟧ : ∀ (e : Expr') → e ↪'* (ℕ→Expr' ⟦ e ⟧')
↪'-⟦⟧ = {!!}
