--------------------------------------------------------------------------------
-- Part 1: Two Languages with Natural Numbers
--------------------------------------------------------------------------------

module Exam.Part1.Specification where

open import Data.Nat using (ℕ; suc; zero; _+_)
open import Relation.Binary.PropositionalEquality using (_≡_; refl; sym; trans; cong; cong₂; subst; module ≡-Reasoning)

-- Fixities for the operators

infix 5 _`+_
infix 6 #_

-- A language with natural number constants and addition.

data Expr : Set where
  #_   : ℕ → Expr
  _`+_ : Expr → Expr → Expr

-- A language with natural number constructors and addition.

data Expr' : Set where
  `zero : Expr'
  `suc  : Expr' → Expr'
  _`+_ : Expr' → Expr' → Expr'

-- Translations between the languages

tr : Expr → Expr'
tr (e₁ `+ e₂) = tr e₁ `+ tr e₂
tr (# zero)   = `zero
tr (# suc x)  = `suc (tr (# x))

tr' : Expr' → Expr
tr' `zero      = # zero
tr' (`suc e)   = (# 1) `+ tr' e
tr' (e₁ `+ e₂) = tr' e₁ `+ tr' e₂

-- Denotational semantics mapping expressions to their values.

⟦_⟧ : Expr → ℕ
⟦ # x ⟧      = x
⟦ e₁ `+ e₂ ⟧ = ⟦ e₁ ⟧ + ⟦ e₂ ⟧

⟦_⟧' : Expr' → ℕ
⟦ `zero ⟧'    = zero
⟦ `suc e ⟧'   = suc ⟦ e ⟧'
⟦ e₁ `+ e₂ ⟧' = ⟦ e₁ ⟧' + ⟦ e₂ ⟧'

-- Small-Step semantics

data Value : Expr → Set where
  val-# : ∀ n → Value (# n)

data Value' : Expr' → Set where
  val-zero : Value' `zero
  val-suc : ∀ {e} → Value' e → Value' (`suc e)

data _↪_ : Expr → Expr → Set where
  β-+  : ∀ {n₁ n₂} →
    ((# n₁) `+ (# n₂)) ↪ (# (n₁ + n₂))
  ξ-+₁ : ∀ {e₁ e₂ e₁'} →
    e₁ ↪ e₁' →
    (e₁ `+ e₂) ↪ (e₁' `+ e₂)
  ξ-+₂ : ∀ {e₁ e₂ e₂'} →
    e₂ ↪ e₂' →
    (e₁ `+ e₂) ↪ (e₁ `+ e₂')

data _↪'_ : Expr' → Expr' → Set where
  β-+-suc  : ∀ {e₁ e₂} →
    (`suc e₁ `+ e₂) ↪' `suc (e₁ `+ e₂)
  β-+-zero : ∀ {e} →
    (`zero `+ e) ↪' e
  ξ-+₁     : ∀ {e₁ e₂ e₁'} →
    e₁ ↪' e₁' →
    (e₁ `+ e₂) ↪' (e₁' `+ e₂)
  ξ-+₂     : ∀ {e₁ e₂ e₂'} →
    e₂ ↪' e₂' →
    (e₁ `+ e₂) ↪' (e₁ `+ e₂')
  ξ-suc    : ∀ {e e'} →
    e ↪' e' →
    (`suc e) ↪' (`suc e')

data _↪*_ : Expr → Expr → Set where
  ↪*-refl : ∀ {e : Expr}
    → e ↪* e
  ↪*-step : ∀ {e₁ e₂ e₃ : Expr}
    → e₁ ↪  e₂
    → e₂ ↪* e₃
    → e₁ ↪* e₃

↪*-trans : ∀ {e₁ e₂ e₃ : Expr}
  → e₁ ↪* e₂
  → e₂ ↪* e₃
  → e₁ ↪* e₃
↪*-trans ↪*-refl       q = q
↪*-trans (↪*-step x p) q = ↪*-step x (↪*-trans p q)

module ↪*-Reasoning where
  infix 1 begin_
  infixr 2 _↪⟨_⟩_ _↪*⟨_⟩_ _≡⟨_⟩_ _≡⟨⟩_
  infix 3 _∎

  begin_ : ∀ {e₁ e₂ : Expr} → e₁ ↪* e₂ → e₁ ↪* e₂
  begin p = p

  _↪⟨_⟩_ : ∀ (e₁ {e₂} {e₃} : Expr) → e₁ ↪ e₂ → e₂ ↪* e₃ → e₁ ↪* e₃
  _ ↪⟨ p ⟩ q = ↪*-step p q

  _↪*⟨_⟩_ : ∀ (e₁ {e₂} {e₃} : Expr) → e₁ ↪* e₂ → e₂ ↪* e₃ → e₁ ↪* e₃
  _ ↪*⟨ p ⟩ q = ↪*-trans p q

  _≡⟨_⟩_ : ∀ (e₁ {e₂} {e₃} : Expr) → e₁ ≡ e₂ → e₂ ↪* e₃ → e₁ ↪* e₃
  _ ≡⟨ refl ⟩ q = q

  _≡⟨⟩_ : ∀ (e₁ {e₂} {e₃} : Expr) → e₁ ↪* e₂ → e₁ ↪* e₂
  _ ≡⟨⟩ q = q

  _∎ : ∀ (e : Expr) → e ↪* e
  _ ∎ = ↪*-refl

data _↪'*_ : Expr' → Expr' → Set where
  ↪'*-refl : ∀ {e : Expr'}
    → e ↪'* e
  ↪'*-step : ∀ {e₁ e₂ e₃ : Expr'}
    → e₁ ↪'  e₂
    → e₂ ↪'* e₃
    → e₁ ↪'* e₃

↪'*-trans : ∀ {e₁ e₂ e₃ : Expr'}
  → e₁ ↪'* e₂
  → e₂ ↪'* e₃
  → e₁ ↪'* e₃
↪'*-trans ↪'*-refl       q = q
↪'*-trans (↪'*-step x p) q = ↪'*-step x (↪'*-trans p q)

module ↪'*-Reasoning where
  infix 1 begin_
  infixr 2 _↪'⟨_⟩_ _↪'*⟨_⟩_ _≡⟨_⟩_ _≡⟨⟩_
  infix 3 _∎

  begin_ : ∀ {e₁ e₂ : Expr'} → e₁ ↪'* e₂ → e₁ ↪'* e₂
  begin p = p

  _↪'⟨_⟩_ : ∀ (e₁ {e₂} {e₃} : Expr') → e₁ ↪' e₂ → e₂ ↪'* e₃ → e₁ ↪'* e₃
  _ ↪'⟨ p ⟩ q = ↪'*-step p q

  _↪'*⟨_⟩_ : ∀ (e₁ {e₂} {e₃} : Expr') → e₁ ↪'* e₂ → e₂ ↪'* e₃ → e₁ ↪'* e₃
  _ ↪'*⟨ p ⟩ q = ↪'*-trans p q

  _≡⟨_⟩_ : ∀ (e₁ {e₂} {e₃} : Expr') → e₁ ≡ e₂ → e₂ ↪'* e₃ → e₁ ↪'* e₃
  _ ≡⟨ refl ⟩ q = q

  _≡⟨⟩_ : ∀ (e₁ {e₂} {e₃} : Expr') → e₁ ↪'* e₂ → e₁ ↪'* e₂
  _ ≡⟨⟩ q = q

  _∎ : ∀ (e : Expr') → e ↪'* e
  _ ∎ = ↪'*-refl

-- Converting a natural number to an Expr' by replacing suc with `suc
-- and zero with `zero.

ℕ→Expr' : ℕ → Expr'
ℕ→Expr' n = tr (# n)
