{-# OPTIONS --allow-unsolved-metas #-}

module Exam.Part3.ConversionProperties where

open import Data.Product renaming (_,_ to ⟨_,_⟩)
open import Data.Sum
open import Data.Nat using (ℕ; zero; suc; _+_)
open import Function using (_∘_)
open import Relation.Binary.PropositionalEquality using (_≡_; refl; sym; subst; cong; cong₂; module ≡-Reasoning)
open import Data.Fin using (zero; suc) renaming (Fin to 𝔽)
open import Relation.Nullary using (¬_)
open import Data.Empty using (⊥; ⊥-elim)

open import Exam.Part3.Specification
open import Exam.Part3.Confluence using (confluence; _↪ₚ_; _↪ₚ*_; ↪ₚ-refl; ↪ₚ-ren; ↪ₚ*→↪*; ↪*→↪ₚ*); open _↪ₚ*_; open _↪ₚ_
open import Exam.Part3.Composition using (swap-0↦-extᵣ-fusion; swap-0↦-extₛ-fusion)

--------------------------------------------------------------------------------

-- Conversion is an equivalence relation

≈-refl : ∀ {n} {e : Term n}
  → e ≈ e
≈-refl = {!!}

≈-sym : ∀ {n} {e₁ e₂ : Term n}
  → e₁ ≈ e₂
  → e₂ ≈ e₁
≈-sym = {!!}

≈-trans : ∀ {n} {e₁ e₂ e₃ : Term n}
  → e₁ ≈ e₂
  → e₂ ≈ e₃
  → e₁ ≈ e₃
≈-trans = {!!}

--------------------------------------------------------------------------------

-- Reduction implies conversion

↪→≈ : ∀ {n} {e₁ e₂ : Term n}
  → e₁ ↪ e₂
  → e₁ ≈ e₂
↪→≈ = {!!}

--------------------------------------------------------------------------------

-- Congruency rules from parallel reduction, but using ↪* instead of ↪

ξₚ*-λ : ∀ {n} {e e' : Term (suc n)}
  → e ↪ₚ* e'
  ---------------
  → `λ e ↪ₚ* `λ e'
ξₚ*-λ = {!!}

ξₚ*-· : ∀ {n} {e₁ e₂ e₁' e₂' : Term n}
  → e₁ ↪ₚ* e₁'
  → e₂ ↪ₚ* e₂'
  ---------------------
  → e₁ · e₂ ↪ₚ* e₁' · e₂'
ξₚ*-· = {!!}

ξₚ*-∀ : ∀ {n} {t₁ t₁' : Term n} {t₂ t₂' : Term (suc n)}
  → t₁ ↪ₚ* t₁'
  → t₂ ↪ₚ* t₂'
  -------------------------------
  → ∀[x⦂ t₁ ] t₂ ↪ₚ* ∀[x⦂ t₁' ] t₂'
ξₚ*-∀ = {!!}

--------------------------------------------------------------------------------
-- "Substituting two convertible terms into another term, yields again convertible terms."

-- A substitution σ₁ parallel-reduces in many steps to σ₂, if each
-- term of σ₁ parallel-reduces in many steps to the corresponding term of σ₂
_↪ₚ*σ_ : ∀ {m n} (σ₁ σ₂ : Sub m n) → Set
σ₁ ↪ₚ*σ σ₂ = ∀ x → σ₁ x ↪ₚ* σ₂ x

-- A substitution σ₁ reduces in many steps to σ₂, if each
-- term of σ₁ reduces in many steps to the corresponding term of σ₂
_↪*σ_ : ∀ {m n} (σ₁ σ₂ : Sub m n) → Set
σ₁ ↪*σ σ₂ = ∀ x → σ₁ x ↪* σ₂ x

-- A substitution σ₁ is convertible to a σ₂, if each
-- term of σ₁ is convertible to the corresponding term of σ₂
_≈σ_ : ∀ {m n} (σ₁ σ₂ : Sub m n) → Set
σ₁ ≈σ σ₂ = ∀ x → σ₁ x ≈ σ₂ x

↪ₚ*-ren :
  ∀ {m n} {t t' : Term m} (ρ : Ren m n) →
  t ↪ₚ* t' →
  ren ρ t ↪ₚ* ren ρ t'
↪ₚ*-ren = {!!}

↪ₚ*σ-ext :  ∀ {m n} {σ σ' : Sub m n}
  → σ ↪ₚ*σ σ'
  → extₛ σ ↪ₚ*σ extₛ σ'
↪ₚ*σ-ext = {!!}

↪ₚ*σ-sub : ∀ {m n} {e : Term m} {σ σ' : Sub m n} →
  σ ↪ₚ*σ σ' →
  sub σ e ↪ₚ* sub σ' e
↪ₚ*σ-sub = {!!}

↪*σ-sub : ∀ {m n} {e : Term m} {σ σ' : Sub m n} →
  σ ↪*σ σ' →
  sub σ e ↪* sub σ' e
↪*σ-sub = {!!}

≈σ-sub : ∀ {m n} {σ σ' : Sub m n} {e : Term m}
  → σ ≈σ σ'
  → sub σ e ≈ sub σ' e
≈σ-sub = {!!}

≈→≈σ : ∀ {n} {e e' : Term n}
  → e ≈ e'
  → 0↦ e ≈σ 0↦ e'
≈→≈σ = {!!}

≈σ-sub₁ : ∀ {n} {e : Term (suc n)} {e₁ e₂ : Term n}
  → e₁ ≈ e₂
  → e [ e₁ ] ≈ e [ e₂ ]
≈σ-sub₁ = {!!}

≈σ-refl : ∀ {m n} {σ : Sub m n} →
  σ ≈σ σ
≈σ-refl = {!!}

--------------------------------------------------------------------------------
-- "Applying a substitution to two convertible terms, yields again convertible terms"

-- Renaming preserves reduction

↪-ren : ∀ {m n} {e e' : Term m} (ρ : Ren m n)
  → e ↪ e'
  → ren ρ e ↪ ren ρ e'
-- HINT: can also be solved with eq chains instead of subst
↪-ren = {!!}

-- Renaming preserves many reduction steps

↪*-ren : ∀ {m n} {e e' : Term m} (ρ : Ren m n)
  → e ↪* e'
  → ren ρ e ↪* ren ρ e'
↪*-ren = {!!}

-- Renaming preserves convertibility

≈-ren : ∀ {m n} {e e' : Term m} (ρ : Ren m n)
  → e ≈ e'
  → ren ρ e ≈ ren ρ e'
≈-ren = {!!}

-- Substitution preserves reduction

↪-sub : ∀ {m n} {e e' : Term m} (σ : Sub m n)
  → e ↪ e'
  → sub σ e ↪ sub σ e'
-- HINT: can also be solved with eq chains instead of subst
↪-sub = {!!}

-- Substitution preserves many reduction steps

↪*-sub : ∀ {m n} {e e' : Term m} (σ : Sub m n)
  → e ↪* e'
  → sub σ e ↪* sub σ e'
↪*-sub = {!!}

-- Substitution preserves convertibility

≈-sub : ∀ {m n} {e e' : Term m} (σ : Sub m n)
  → e ≈ e'
  → sub σ e ≈ sub σ e'
≈-sub = {!!}

--------------------------------------------------------------------------------
-- "Typing is preserved along convertible types in the context"

-- Two contexts Γ₁ and Γ₂ are convertible, if each type in Γ₁ is
-- convertible to the corresponding type in Γ₂.

_≈ᶜ_ : ∀ {n} → Context n → Context n → Set
Γ₁ ≈ᶜ Γ₂ = ∀ x → lookup Γ₁ x ≈ lookup Γ₂ x

≈ᶜ-refl : ∀ {n} {Γ : Context n}
  → Γ ≈ᶜ Γ
≈ᶜ-refl = {!!}

≈-ext : ∀ {n} {Γ₁ Γ₂ : Context n} {t₁ t₂ : Term n}
  → Γ₁ ≈ᶜ Γ₂
  → t₁ ≈ t₂
  → (Γ₁ , t₁) ≈ᶜ (Γ₂ , t₂)
≈-ext = {!!}

≈-Γ-⊢ : ∀ {n} {Γ₁ Γ₂ : Context n} {e t : Term n}
  → Γ₁ ≈ᶜ Γ₂
  → Γ₁ ⊢ e ⦂ t
  → Γ₂ ⊢ e ⦂ t
≈-Γ-⊢ = {!!}

≈-Γ-⊢₁ : ∀ {n} {Γ : Context n} {t₁ t₂ : Term n} {e t : Term (suc n)}
  → t₁ ≈ t₂
  → Γ , t₁ ⊢ e ⦂ t
  → Γ , t₂ ⊢ e ⦂ t
≈-Γ-⊢₁ = {!!}
